@@ -2,13 +2,11 @@ use glib::subclass::prelude::*;
 use gtk::{gio, glib};
 
 mod imp {
-    use ashpd::desktop::open_uri::OpenDirectoryRequest;
     use std::{
         cell::{Cell, RefCell},
         ffi::{OsStr, OsString},
-        fs::File,
         marker::PhantomData,
-        os::{fd::AsFd, unix::prelude::OsStringExt},
+        os::unix::prelude::OsStringExt,
         path::{Component, Path, PathBuf},
         str,
         time::Duration,
@@ -18,12 +16,15 @@ mod imp {
     use futures_util::future::{abortable, FutureExt};
     use gettextrs::*;
     use glib::{debug, error, warn, Properties};
-    use gtk::{gdk, gio, glib, CompositeTemplate};
+    use gtk::{
+        gdk::{self, Key, ModifierType},
+        gio, glib, CompositeTemplate,
+    };
 
     use crate::{
         config::{self, G_LOG_DOMAIN},
         parse::{self, time_to_entry_text},
-        util::gettext_f,
+        util::{gettext_f, with_recursive_children},
         video_preview::VtVideoPreview,
     };
 
@@ -102,6 +103,12 @@ mod imp {
         overlay_error_page: TemplateChild<adw::ToastOverlay>,
         #[template_child]
         toolbar_view: TemplateChild<adw::ToolbarView>,
+        #[template_child]
+        popover_options: TemplateChild<gtk::Popover>,
+        #[template_child]
+        switch_row_reencode: TemplateChild<adw::SwitchRow>,
+        #[template_child]
+        switch_row_remove_audio: TemplateChild<adw::SwitchRow>,
 
         #[property(get = Self::is_playing, set = Self::set_is_playing, explicit_notify)]
         is_playing: PhantomData<bool>,
@@ -312,131 +319,93 @@ Please choose another file.",
 
             self.video_preview.pause();
 
-            let obj = self.obj().clone();
+            let output_path = self
+                .output_file
+                .borrow()
+                .as_ref()
+                .and_then(|file| file.path())
+                .unwrap_or_else(|| {
+                    let document_portal_components = [
+                        Component::RootDir,
+                        Component::Normal(OsStr::new("run")),
+                        Component::Normal(OsStr::new("user")),
+                        Component::Normal(OsStr::new("doc")),
+                    ];
 
-            let future = async move {
-                let imp = obj.imp();
+                    let mut components = input_path.components();
 
-                let output_path = imp
-                    .output_file
-                    .borrow()
-                    .as_ref()
-                    .and_then(|file| file.path())
-                    .unwrap_or_else(|| {
-                        let document_portal_components = [
-                            Component::RootDir,
-                            Component::Normal(OsStr::new("run")),
-                            Component::Normal(OsStr::new("user")),
-                            Component::Normal(OsStr::new("doc")),
-                        ];
+                    let prefix = if components.next() == Some(document_portal_components[0])
+                        && components.next() == Some(document_portal_components[1])
+                        && components.next() == Some(document_portal_components[2])
+                        && components.nth(1) == Some(document_portal_components[3])
+                    {
+                        // input_path comes from the document portal, no use in opening the
+                        // file chooser there.
+                        None
+                    } else {
+                        input_path.parent().map(Path::to_path_buf)
+                    };
 
-                        let mut components = input_path.components();
+                    let mut path = prefix.unwrap_or_default();
 
-                        let prefix = if components.next() == Some(document_portal_components[0])
-                            && components.next() == Some(document_portal_components[1])
-                            && components.next() == Some(document_portal_components[2])
-                            && components.nth(1) == Some(document_portal_components[3])
-                        {
-                            // input_path comes from the document portal, no use in opening the
-                            // file chooser there.
-                            None
-                        } else {
-                            input_path.parent().map(Path::to_path_buf)
-                        };
+                    path.push(format!(
+                        "{}{}.{}",
+                        input_path.file_stem().and_then(OsStr::to_str).unwrap_or(""),
+                        // Translators: this is appended to the output video file name.
+                        // So for example "my video.mp4" will become "my video (trimmed).mp4".
+                        gettext(" (trimmed)"),
+                        extension
+                    ));
 
-                        let mut path = prefix.unwrap_or_default();
-
-                        path.push(format!(
-                            "{}{}.{}",
-                            input_path.file_stem().and_then(OsStr::to_str).unwrap_or(""),
-                            // Translators: this is appended to the output video file name.
-                            // So for example "my video.mp4" will become "my video (trimmed).mp4".
-                            gettext(" (trimmed)"),
-                            extension
-                        ));
-
-                        path
-                    });
-
-                // FIXME: migrate to FileDialog once the UI is redesigned to put the checkboxes
-                // into the UI itself, rather than into the save dialog.
-                let file_chooser = gtk::FileChooserNative::builder()
-                    .transient_for(&obj)
-                    .action(gtk::FileChooserAction::Save)
-                    .modal(true)
-                    .build();
-                if let Some(parent) = output_path.parent() {
-                    if parent.to_str().map(|x| !x.is_empty()).unwrap_or(false) {
-                        debug!("setting current folder to {:?}", parent);
-                        let _ = file_chooser.set_current_folder(Some(&gio::File::for_path(parent)));
-                    }
-                }
-                if let Some(name) = output_path.file_name().and_then(OsStr::to_str) {
-                    debug!("setting current name to {:?}", name);
-                    file_chooser.set_current_name(name);
-                }
-
-                // Translators: checkbox in output file selection dialog that strips audio from the
-                // video file.
-                file_chooser.add_choice("no-audio", gettext("Remove audio"), &[]);
-
-                file_chooser.add_choice(
-                    "reencode",
-                    // Translators: checkbox in output file selection dialog.
-                    gettext("Accurate trimming, but slower and may lose quality"),
-                    &[],
-                );
-
-                let (tx, rx) = futures_channel::oneshot::channel();
-
-                let tx = RefCell::new(Some(tx));
-                file_chooser.connect_response({
-                    let obj = obj.downgrade();
-                    move |file_chooser, response| {
-                        if let Some(tx) = tx.borrow_mut().take() {
-                            if response == gtk::ResponseType::Accept {
-                                if let Some(path) = file_chooser.file().and_then(|file| file.path())
-                                {
-                                    let no_audio = file_chooser
-                                        .choice("no-audio")
-                                        .map(|choice| choice == "true")
-                                        .unwrap_or(false);
-                                    let reencode = file_chooser
-                                        .choice("reencode")
-                                        .map(|choice| choice == "true")
-                                        .unwrap_or(false);
-                                    tx.send(Some((path, no_audio, reencode))).unwrap();
-                                } else {
-                                    let dialog = adw::AlertDialog::builder()
-                                        // Translators: error dialog title.
-                                        .heading(gettext("Error"))
-                                        .body(gettext(
-                                            // Translators: error dialog text.
-                                            "Video Trimmer can only operate on local files. \
-Please choose another file.",
-                                        ))
-                                        .build();
-                                    // Translators: error dialog button.
-                                    dialog.add_response("ok", &gettext("_OK"));
-                                    dialog.present(Some(&obj.upgrade().unwrap()));
-
-                                    tx.send(None).unwrap();
-                                }
-                            } else {
-                                tx.send(None).unwrap();
-                            }
-                        }
-                    }
+                    path
                 });
 
-                file_chooser.show();
+            let file_dialog = gtk::FileDialog::builder().modal(true).build();
+            if let Some(parent) = output_path.parent() {
+                if parent.to_str().map(|x| !x.is_empty()).unwrap_or(false) {
+                    debug!("setting initial folder to {:?}", parent);
+                    file_dialog.set_initial_folder(Some(&gio::File::for_path(parent)));
+                }
+            }
+            if let Some(name) = output_path.file_name().and_then(OsStr::to_str) {
+                debug!("setting initial name to {:?}", name);
+                file_dialog.set_initial_name(Some(name));
+            }
 
-                let (output_path, no_audio, reencode) = if let Some(value) = rx.await.unwrap() {
-                    value
-                } else {
-                    return;
+            let obj = self.obj().clone();
+            let reencode = self.switch_row_reencode.is_active();
+            let no_audio = self.switch_row_remove_audio.is_active();
+
+            let future = async move {
+                let output_path = match file_dialog.save_future(Some(&obj)).await {
+                    Ok(file) => {
+                        if let Some(path) = file.path() {
+                            path
+                        } else {
+                            let dialog = adw::AlertDialog::builder()
+                                // Translators: error dialog title.
+                                .heading(gettext("Error"))
+                                .body(gettext(
+                                    // Translators: error dialog text.
+                                    "Video Trimmer can only operate on local files. \
+Please choose another file.",
+                                ))
+                                .build();
+                            // Translators: error dialog button.
+                            dialog.add_response("ok", &gettext("_OK"));
+                            dialog.present(Some(&obj));
+                            return;
+                        }
+                    }
+                    Err(err) => {
+                        if !err.matches(gtk::DialogError::Dismissed) {
+                            warn!("file dialog error: {err:?}");
+                        }
+                        return;
+                    }
                 };
 
+                let imp = obj.imp();
                 imp.do_trim(&input_path, output_path, no_audio, reencode, start, end);
             };
 
@@ -733,8 +702,6 @@ Please choose another file.",
                 "ffprobe".as_ref(),
                 "-print_format".as_ref(),
                 "json".as_ref(),
-                "-select_streams".as_ref(),
-                "a".as_ref(),
                 "-show_streams".as_ref(),
                 input_path.as_ref().unwrap().as_ref(),
             ];
@@ -763,9 +730,17 @@ Please choose another file.",
                                     let imp = obj.imp();
                                     let output = json::parse(stdout)
                                         .expect("ffprobe should return valid JSON");
-                                    let mut audio_formats = output["streams"]
+                                    let streams = &output["streams"];
+
+                                    let mut audio_formats = streams
                                         .members()
-                                        .filter_map(|stream| stream["codec_name"].as_str())
+                                        .filter_map(|stream| {
+                                            if stream["codec_type"].as_str() == Some("audio") {
+                                                stream["codec_name"].as_str()
+                                            } else {
+                                                None
+                                            }
+                                        })
                                         .inspect(|name| debug!("audio codec: {}", name));
 
                                     // Some Sony cameras produce .mp4 videos with PCM audio. This is invalid
@@ -778,6 +753,39 @@ Please choose another file.",
                                         );
                                         imp.do_not_default_to_mp4.set(true);
                                     }
+
+                                    // Only get the first one we can find, because that's what GTK is going to display
+                                    let frame_rate_fraction = streams
+                                        .members()
+                                        .find(|stream| {
+                                            stream["codec_type"].as_str() == Some("video")
+                                        })
+                                        .and_then(|video_stream| {
+                                            video_stream["r_frame_rate"].as_str()
+                                        })
+                                        .inspect(|value| debug!("r_frame_rate: {value}"))
+                                        .and_then(|frame_rate| frame_rate.split_once('/'));
+
+                                    let mut frame_time = None;
+                                    if let Some((numerator, denominator)) = frame_rate_fraction {
+                                        if let Some((numerator, denominator)) = Option::zip(
+                                            numerator.parse::<f64>().ok(),
+                                            denominator.parse::<f64>().ok(),
+                                        ) {
+                                            if numerator > 0. && denominator > 0. {
+                                                frame_time = Some(Duration::from_secs_f64(
+                                                    denominator / numerator,
+                                                ));
+                                            }
+                                        }
+                                    }
+
+                                    if let Some(frame_time) = frame_time {
+                                        debug!("computed frame time: {frame_time:?}");
+                                        imp.video_preview.get().set_frame_time_approx(frame_time);
+                                    } else {
+                                        warn!("failed get frame time, stepping will not work")
+                                    }
                                 }
                                 Err(err) => {
                                     // ffmpeg's JSON output fixes up invalid UTF-8 for us with
@@ -800,6 +808,14 @@ Please choose another file.",
             };
             glib::MainContext::default().spawn_local(future);
         }
+
+        pub fn step_forward(&self) {
+            self.video_preview.step_forward()
+        }
+
+        pub fn step_back(&self) {
+            self.video_preview.step_back()
+        }
     }
 
     #[glib::object_subclass]
@@ -812,6 +828,14 @@ Please choose another file.",
             Self::bind_template(klass);
 
             klass.install_property_action("win.play-pause", "is-playing");
+
+            klass.install_action("win.step-forward", None, |window, _, _| {
+                window.imp().step_forward()
+            });
+            klass.install_action("win.step-back", None, |window, _, _| {
+                window.imp().step_back()
+            });
+
             klass.install_action("win.close", None, |window, _, _| window.close());
 
             klass.install_action("win.trim", None, |window, _, _| {
@@ -825,7 +849,7 @@ Please choose another file.",
             klass.install_action("win.about", None, |window, _, _| {
                 let resource_path = "/org/gnome/gitlab/YaLTeR/VideoTrimmer/\
                                      org.gnome.gitlab.YaLTeR.VideoTrimmer.metainfo.xml";
-                let about_window = adw::AboutDialog::from_appdata(resource_path, Some("0.9.0"));
+                let about_window = adw::AboutDialog::from_appdata(resource_path, Some("25.03"));
                 about_window.set_version(config::VERSION);
                 // Translators: shown in the About dialog, put your name here.
                 about_window.set_translator_credits(&gettext("translator-credits"));
@@ -834,17 +858,23 @@ Please choose another file.",
                     &gettext("Contribute Translations"),
                     "https://l10n.gnome.org/module/video-trimmer/",
                 );
+                about_window.add_other_app(
+                    "org.gnome.gitlab.YaLTeR.Identity",
+                    // Translators: name of https://gitlab.gnome.org/YaLTeR/identity
+                    &gettext("Identity"),
+                    // Translators: summary of https://gitlab.gnome.org/YaLTeR/identity
+                    &gettext("Compare images and videos"),
+                );
                 about_window.present(Some(window));
 
                 // DL doesn't extract release notes from metainfo, so let's help it out with the
                 // ones shown in the dialog.
                 let gettext = |_| ();
-                gettext("This release improves the behavior of shortcuts and does a minor visual refresh for GNOME 47.");
-                gettext("Added I and O shortcuts to set start and end trimming point.");
-                gettext("Made all single-letter shortcuts work even when the time input fields are focused.");
-                gettext("Changed the play/pause Ctrl+Space shortcut to just Space.");
-                gettext("The video now starts paused after opening.");
-                gettext("Updated to the GNOME 47 platform.");
+                gettext("This release makes the trimming options more discoverable.");
+                gettext("Changed the versioning scheme to year.month.");
+                gettext("Moved the accurate trimming and remove audio options from the save dialog to the main window.");
+                gettext("Added , and . shortcuts for short video stepping.");
+                gettext("Updated to the GNOME 48 platform.");
                 gettext("Updated translations.");
             });
 
@@ -854,32 +884,17 @@ Please choose another file.",
                 |window, _, path| {
                     let path = Vec::<u8>::from_variant(path.unwrap()).unwrap();
                     let path = PathBuf::from(OsString::from_vec(path));
-                    let file = match File::open(path) {
-                        Ok(value) => value,
-                        Err(err) => {
-                            warn!("couldn't open the output file: {:?}", err);
-                            return;
-                        }
-                    };
+                    let file = gio::File::for_path(path);
 
-                    let native = if let Some(value) = window.native() {
-                        value
-                    } else {
-                        warn!("window.native() returned None");
-                        return;
-                    };
-
-                    let future = async move {
-                        let identifier = ashpd::WindowIdentifier::from_native(&native).await;
-                        if let Err(err) = OpenDirectoryRequest::default()
-                            .identifier(identifier)
-                            .send(&file.as_fd())
-                            .await
-                        {
-                            warn!("OpenDirectory returned an error: {:?}", err);
-                        }
-                    };
-                    glib::MainContext::default().spawn_local(future);
+                    gtk::FileLauncher::new(Some(&file)).open_containing_folder(
+                        Some(window),
+                        gio::Cancellable::NONE,
+                        move |res| {
+                            if let Err(err) = res {
+                                warn!("OpenDirectory returned an error: {:?}", err);
+                            }
+                        },
+                    );
                 },
             );
 
@@ -889,6 +904,11 @@ Please choose another file.",
             klass.install_action("win.set-end-as-position", None, |window, _, _| {
                 window.imp().video_preview.set_end_as_position()
             });
+
+            // Add these here instead of set_accels_for_action so that they don't override typing in
+            // the time entries.
+            klass.add_binding_action(Key::period, ModifierType::empty(), "win.step-forward");
+            klass.add_binding_action(Key::comma, ModifierType::empty(), "win.step-back");
         }
 
         fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
@@ -1078,6 +1098,16 @@ Please choose another file.",
                 }
             });
             self.stack.add_controller(drop_target);
+
+            // HACK: Make options popover action row subtitles wrap eagerly to ensure that they fit
+            // into the mobile window widths even with especially long translations, etc.
+            with_recursive_children(self.popover_options.upcast_ref(), &mut |widget| {
+                if let Some(label) = widget.downcast_ref::<gtk::Label>() {
+                    if label.has_css_class("subtitle") {
+                        label.set_max_width_chars(20);
+                    }
+                }
+            });
         }
     }
 
@@ -1115,7 +1145,7 @@ Please choose another file.",
 
 glib::wrapper! {
     pub struct VtWindow(ObjectSubclass<imp::VtWindow>)
-        @extends gtk::Widget, gtk::Window, gtk::ApplicationWindow,
+        @extends gtk::Widget, gtk::Window, gtk::ApplicationWindow, adw::ApplicationWindow,
         @implements gio::ActionMap, gio::ActionGroup;
 }
 
